from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import str
import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from future import standard_library
standard_library.install_aliases()
from builtins import object
import io
import json
import os
from enum import Enum
from typing import Any
from typing import Dict
from typing import Optional
from typing import Union

from gcloud.rest.auth import SyncSession  # pylint: disable=no-name-in-module
from gcloud.rest.auth import BUILD_GCLOUD_REST  # pylint: disable=no-name-in-module
from gcloud.rest.auth import Token  # pylint: disable=no-name-in-module

# Selectively load libraries based on the package
if BUILD_GCLOUD_REST:
    from requests import Session
else:
    from aiohttp import ClientSession as Session  # type: ignore[no-redef]


API_ROOT = 'https://www.googleapis.com/bigquery/v2'
SCOPES = [
    'https://www.googleapis.com/auth/bigquery.insertdata',
    'https://www.googleapis.com/auth/bigquery',
]

BIGQUERY_EMULATOR_HOST = os.environ.get('BIGQUERY_EMULATOR_HOST')
if BIGQUERY_EMULATOR_HOST:
    API_ROOT = 'http://{}/bigquery/v2'.format((BIGQUERY_EMULATOR_HOST))


class SourceFormat(Enum):
    AVRO = 'AVRO'
    CSV = 'CSV'
    DATASTORE_BACKUP = 'DATASTORE_BACKUP'
    NEWLINE_DELIMITED_JSON = 'NEWLINE_DELIMITED_JSON'
    ORC = 'ORC'
    PARQUET = 'PARQUET'


class Disposition(Enum):
    WRITE_APPEND = 'WRITE_APPEND'
    WRITE_EMPTY = 'WRITE_EMPTY'
    WRITE_TRUNCATE = 'WRITE_TRUNCATE'


class SchemaUpdateOption(Enum):
    ALLOW_FIELD_ADDITION = 'ALLOW_FIELD_ADDITION'
    ALLOW_FIELD_RELAXATION = 'ALLOW_FIELD_RELAXATION'


class BigqueryBase(object):
    def __init__(self,
                 project                = None,
                 service_file                                  = None,
                 session                    = None,
                 token                  = None)        :
        self._project = project
        self.session = SyncSession(session)
        self.token = token or Token(service_file=service_file, scopes=SCOPES,
                                    session=self.session.session)

    def project(self)       :
        if self._project:
            return self._project

        if BIGQUERY_EMULATOR_HOST:
            self._project = str(os.environ.get('BIGQUERY_PROJECT_ID', 'dev'))
            return self._project

        self._project = self.token.get_project()
        if self._project:
            return self._project

        raise Exception('could not determine project, please set it manually')

    def headers(self)                  :
        if BIGQUERY_EMULATOR_HOST:
            return {}

        token = self.token.get()
        return {
            'Authorization': 'Bearer {}'.format((token)),
        }

    def _post_json(
            self, url     , body                , session                   ,
            timeout     )                  :
        payload = json.dumps(body).encode('utf-8')

        headers = self.headers()
        headers.update({
            'Content-Length': str(len(payload)),
            'Content-Type': 'application/json',
        })

        s = SyncSession(session) if session else self.session
        resp = s.post(url, data=payload, headers=headers, params=None,
                            timeout=timeout)
        data                 = resp.json()
        return data

    def _get_url(
            self, url     , session                   ,
            timeout     ,
            params                           = None)                  :
        headers = self.headers()

        s = SyncSession(session) if session else self.session
        resp = s.get(url, headers=headers, timeout=timeout,
                           params=params or {})
        data                 = resp.json()
        return data

    def close(self)        :
        self.session.close()

    def __enter__(self)                  :
        return self

    def __exit__(self, *args     )        :
        self.close()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import zip
import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from future import standard_library
standard_library.install_aliases()
import datetime
import decimal
import logging
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional


log = logging.getLogger(__name__)


try:
    utc = datetime.timezone.utc
except AttributeError:
    # build our own UTC for Python 2
    class UTC(datetime.tzinfo):
        def utcoffset(self,
                      _dt                             )                      :
            return datetime.timedelta(0)

        def tzname(self, _dt                             )       :
            return 'UTC'

        def dst(self, _dt                             )                      :
            return datetime.timedelta(0)

    utc = UTC()  # type: ignore[assignment]


def flatten(x     )       :
    """
    Flatten response objects into something we can actually work with.

    The API returns data of the form:

        {'f': [{'v': ...}]}

    to indicate groupings of fields and their potential for having multiple
    values. We want those to just be plain old objects.
    """
    if isinstance(x, dict):
        # TODO: what if a user has stored a dictionary in their table and that
        # dictionary is shaped the same way as Google's response format?
        if 'f' in x:
            return [flatten(y['v']) for y in x['f']]

        if 'v' in x:
            return flatten(x['v'])

    if isinstance(x, list):
        return [flatten(y) for y in x]

    return x


def parse(field                , value     )       :
    """
    Parse a given field back to a Python object.

    This is often trivial: convert the value from a string to the type
    specified in the field's schema. There's a couple caveats we've identified
    so far, though:

    * NULLABLE fields should be handled specially, eg. so as not to
      accidentally convert them to the schema type.
    * REPEATED fields are nested a biot differently than expected, so we need
      to flatten *first*, then convert.

    `Field = Dict[str, Union[str, 'Field']]`, but wow is that difficult to
    represent in a backwards-enough compatible fashion.
    """
    try:
        convert                       = {  # type: ignore[assignment]
            'BIGNUMERIC': lambda x: decimal.Decimal(
                x, decimal.Context(prec=77)),
            'BOOLEAN': lambda x: x == 'true',
            'BYTES': bytes,
            'FLOAT': float,
            'INTEGER': int,
            'NUMERIC': lambda x: decimal.Decimal(
                x, decimal.Context(prec=38)),
            'RECORD': dict,
            'STRING': str,
            'TIMESTAMP': lambda x: datetime.datetime.fromtimestamp(
                float(x), tz=utc),
        }[field['type']]
    except KeyError:
        # TODO: determine the proper methods for converting the following:
        # DATE -> datetime?
        # DATETIME -> datetime?
        # GEOGRAPHY -> ??
        # TIME -> datetime?
        log.error('Unsupported field type %s. Please open a bug report with '
                  'the following data: %s, %s', field['type'], field['mode'],
                  flatten(value))
        raise

    if field['mode'] == 'NULLABLE' and value is None:
        return value

    if field['mode'] == 'REPEATED':
        if field['type'] == 'RECORD':
            return [{f['name']: parse(f, x)
                     for f, x in zip(field['fields'], xs)}
                    for xs in flatten(value)]

        return [convert(x) for x in flatten(value)]

    if field['type'] == 'RECORD':
        return {f['name']: parse(f, x)
                for f, x in zip(field['fields'], flatten(value))}

    return convert(flatten(value))


def query_response_to_dict(response                )                        :
    """
    Convert a query response to a dictionary.

    API responses for job queries are packed into a difficult-to-use format.
    This method deserializes a response into a List of rows, with each row
    being a dictionary of field names to the row's value.

    This method also handles converting the values according to the schema
    defined in the response (eg. into builtin python types).
    """
    fields = response['schema'].get('fields', [])
    rows = [x['f'] for x in response.get('rows', [])]
    return [{k['name']: parse(k, v) for k, v in zip(fields, row)}
            for row in rows]



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import super
import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from future import standard_library
standard_library.install_aliases()
import io
from typing import Any
from typing import Dict
from typing import Optional
from typing import Union

from gcloud.rest.auth import BUILD_GCLOUD_REST  # pylint: disable=no-name-in-module
from gcloud.rest.auth import Token  # pylint: disable=no-name-in-module

from .bigquery import API_ROOT
from .bigquery import BigqueryBase
from .bigquery import Disposition

# Selectively load libraries based on the package
if BUILD_GCLOUD_REST:
    from requests import Session
else:
    from aiohttp import ClientSession as Session  # type: ignore[no-redef]


class Job(BigqueryBase):
    def __init__(self, job_id                = None,
                 project                = None,
                 service_file                                  = None,
                 session                    = None,
                 token                  = None)        :
        self.job_id = job_id
        super().__init__(project=project, service_file=service_file,
                         session=session, token=token)

    @staticmethod
    def _make_query_body(
            query     ,
            write_disposition             ,
            use_query_cache      ,
            dry_run      , use_legacy_sql      ,
            destination_table               )                  :
        return {
            'configuration': {
                'query': {
                    'query': query,
                    'writeDisposition': write_disposition.value,
                    'destinationTable': {
                        'projectId': destination_table.project,
                        'datasetId': destination_table.dataset_name,
                        'tableId': destination_table.table_name,
                    } if destination_table else destination_table,
                    'useQueryCache': use_query_cache,
                    'useLegacySql': use_legacy_sql,
                },
                'dryRun': dry_run,
            },
        }

    # https://cloud.google.com/bigquery/docs/reference/rest/v2/jobs/get
    def get_job(self, session                    = None,
                      timeout      = 60)                  :
        """Get the specified job resource by job ID."""

        project = self.project()
        url = '{}/projects/{}/jobs/{}'.format((API_ROOT), (project), (self.job_id))

        return self._get_url(url, session, timeout)

    # https://cloud.google.com/bigquery/docs/reference/rest/v2/jobs/getQueryResults
    def get_query_results(self, session                    = None,
                                timeout      = 60,
                                params                           = None,
                                )                  :
        """Get the specified jobQueryResults by job ID."""

        project = self.project()
        url = '{}/projects/{}/queries/{}'.format((API_ROOT), (project), (self.job_id))

        return self._get_url(url, session, timeout, params=params)

    # https://cloud.google.com/bigquery/docs/reference/rest/v2/jobs/cancel
    def cancel(self, session                    = None,
                     timeout      = 60)                  :
        """Cancel the specified job by job ID."""

        project = self.project()
        url = '{}/projects/{}/queries/{}/cancel'.format((API_ROOT), (project), (self.job_id))

        return self._post_json(url, {}, session, timeout)

    # https://cloud.google.com/bigquery/docs/reference/rest/v2/jobs/query
    def query(self, query_request                ,
                    session                    = None,
                    timeout      = 60)                  :
        """Runs a query synchronously and returns query results if completes
        within a specified timeout."""
        project = self.project()
        url = '{}/projects/{}/queries'.format((API_ROOT), (project))

        return self._post_json(url, query_request, session, timeout)

    # https://cloud.google.com/bigquery/docs/reference/rest/v2/jobs/insert
    def insert(self, job                ,
                     session                    = None,
                     timeout      = 60)                  :
        """Insert a new asynchronous job."""
        project = self.project()
        url = '{}/projects/{}/jobs'.format((API_ROOT), (project))

        response = self._post_json(url, job, session, timeout)
        if response['jobReference'].get('jobId'):
            self.job_id = response['jobReference']['jobId']
        return response

    # https://cloud.google.com/bigquery/docs/reference/rest/v2/jobs/insert
    # https://cloud.google.com/bigquery/docs/reference/rest/v2/Job#JobConfigurationQuery
    def insert_via_query(
            self, query     , session                    = None,
            write_disposition              = Disposition.WRITE_EMPTY,
            timeout      = 60, use_query_cache       = True,
            dry_run       = False, use_legacy_sql       = True,
            destination_table                = None)                  :
        """Create table as a result of the query"""
        project = self.project()
        url = '{}/projects/{}/jobs'.format((API_ROOT), (project))

        body = self._make_query_body(query=query,
                                     write_disposition=write_disposition,
                                     use_query_cache=use_query_cache,
                                     dry_run=dry_run,
                                     use_legacy_sql=use_legacy_sql,
                                     destination_table=destination_table)
        response = self._post_json(url, body, session, timeout)
        if not dry_run:
            self.job_id = response['jobReference']['jobId']
        return response

    def result(self,
                     session                    = None)                  :
        data = self.get_job(session)
        status = data.get('status', {})
        if status.get('state') == 'DONE':
            if 'errorResult' in status:
                raise Exception('Job finished with errors', status['errors'])
            return data

        raise OSError('Job results are still pending')



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import super
import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from future import standard_library
standard_library.install_aliases()
import io
from typing import Any
from typing import Dict
from typing import Optional
from typing import Union

from gcloud.rest.auth import BUILD_GCLOUD_REST  # pylint: disable=no-name-in-module
from gcloud.rest.auth import Token  # pylint: disable=no-name-in-module

from .bigquery import API_ROOT
from .bigquery import BigqueryBase

# Selectively load libraries based on the package
if BUILD_GCLOUD_REST:
    from requests import Session
else:
    from aiohttp import ClientSession as Session  # type: ignore[no-redef]


class Dataset(BigqueryBase):
    def __init__(self, dataset_name                = None,
                 project                = None,
                 service_file                                  = None,
                 session                    = None,
                 token                  = None)        :
        self.dataset_name = dataset_name
        super().__init__(project=project, service_file=service_file,
                         session=session, token=token)

    # https://cloud.google.com/bigquery/docs/reference/rest/v2/tables/list
    def list_tables(
            self, session                    = None,
            timeout      = 60,
            params                           = None)                  :
        """List tables in a dataset."""
        project = self.project()
        if not self.dataset_name:
            raise ValueError('could not determine dataset,'
                             ' please set it manually')

        url = ('{}/projects/{}/datasets/'
               '{}/tables'.format((API_ROOT), (project), (self.dataset_name)))
        return self._get_url(url, session, timeout, params=params)

    # https://cloud.google.com/bigquery/docs/reference/rest/v2/datasets/list
    def list_datasets(
            self, session                    = None,
            timeout      = 60,
            params                           = None)                  :
        """List datasets in current project."""
        project = self.project()

        url = ('{}/projects/{}/datasets'.format((API_ROOT), (project)))
        return self._get_url(url, session, timeout, params=params)

    # https://cloud.google.com/bigquery/docs/reference/rest/v2/datasets/get
    def get(self, session                    = None,
                  timeout      = 60,
                  params                           = None)                  :
        """Get a specific dataset in current project."""
        project = self.project()
        if not self.dataset_name:
            raise ValueError('could not determine dataset,'
                             ' please set it manually')

        url = ('{}/projects/{}/datasets/'
               '{}'.format((API_ROOT), (project), (self.dataset_name)))
        return self._get_url(url, session, timeout, params=params)

    # https://cloud.google.com/bigquery/docs/reference/rest/v2/datasets/insert
    def insert(self, dataset                ,
                     session                    = None,
                     timeout      = 60)                  :
        """Create datasets in current project."""
        project = self.project()

        url = ('{}/projects/{}/datasets'.format((API_ROOT), (project)))
        return self._post_json(url, dataset, session, timeout)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gcloud', 'gcloud.rest', 'gcloud.rest.bigquery']

package_data = \
{'': ['*']}

install_requires = \
['gcloud-rest-auth>=3.1.0,<5.0.0']

setup_kwargs = {
    'name': 'gcloud-rest-bigquery',
    'version': '6.1.2',
    'description': 'Python Client for Google Cloud BigQuery',
    'long_description': "(Asyncio OR Threadsafe) Python Client for Google Cloud BigQuery\n===============================================================\n\n    This is a shared codebase for ``gcloud-rest-bigquery`` and\n    ``gcloud-rest-bigquery``\n\n|pypi| |pythons-aio| |pythons-rest|\n\nInstallation\n------------\n\n.. code-block:: console\n\n    $ pip install --upgrade gcloud-{aio,rest}-bigquery\n\nUsage\n-----\n\nWe're still working on documentation -- for now, you can use the `smoke test`_\nas an example.\n\nEmulators\n~~~~~~~~~\n\nFor testing purposes, you may want to use ``gcloud-rest-bigquery`` along with a\nlocal emulator. Setting the ``$BIGQUERY_EMULATOR_HOST`` environment variable\nto the address of your emulator should be enough to do the trick.\n\nContributing\n------------\n\nPlease see our `contributing guide`_.\n\n.. _contributing guide: https://github.com/talkiq/gcloud-rest/blob/master/.github/CONTRIBUTING.rst\n.. _smoke test: https://github.com/talkiq/gcloud-rest/blob/master/bigquery/tests/integration/smoke_test.py\n\n.. |pypi| image:: https://img.shields.io/pypi/v/gcloud-rest-bigquery.svg?style=flat-square\n    :alt: Latest PyPI Version (gcloud-rest-bigquery)\n    :target: https://pypi.org/project/gcloud-rest-bigquery/\n\n.. |pythons-aio| image:: https://img.shields.io/pypi/pyversions/gcloud-rest-bigquery.svg?style=flat-square&label=python (aio)\n    :alt: Python Version Support (gcloud-rest-bigquery)\n    :target: https://pypi.org/project/gcloud-rest-bigquery/\n\n.. |pythons-rest| image:: https://img.shields.io/pypi/pyversions/gcloud-rest-bigquery.svg?style=flat-square&label=python (rest)\n    :alt: Python Version Support (gcloud-rest-bigquery)\n    :target: https://pypi.org/project/gcloud-rest-bigquery/\n",
    'author': 'Vi Engineering',
    'author_email': 'voiceai-eng@dialpad.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/talkiq/gcloud-aio',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*',
}


setup(**setup_kwargs)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import super
import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from future import standard_library
standard_library.install_aliases()
import io
import json
import uuid
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from gcloud.rest.auth import SyncSession  # pylint: disable=no-name-in-module
from gcloud.rest.auth import BUILD_GCLOUD_REST  # pylint: disable=no-name-in-module
from gcloud.rest.auth import Token  # pylint: disable=no-name-in-module

from .bigquery import API_ROOT
from .bigquery import BigqueryBase
from .bigquery import Disposition
from .bigquery import SchemaUpdateOption
from .bigquery import SourceFormat
from .job import Job

# Selectively load libraries based on the package
if BUILD_GCLOUD_REST:
    from requests import Session
else:
    from aiohttp import ClientSession as Session  # type: ignore[no-redef]


class Table(BigqueryBase):
    def __init__(self, dataset_name     , table_name     ,
                 project                = None,
                 service_file                                  = None,
                 session                    = None,
                 token                  = None)        :
        self.dataset_name = dataset_name
        self.table_name = table_name
        super().__init__(project=project, service_file=service_file,
                         session=session, token=token)

    @staticmethod
    def _mk_unique_insert_id(row                )       :
        # pylint: disable=unused-argument
        return uuid.uuid4().hex

    def _make_copy_body(
            self, source_project     , destination_project     ,
            destination_dataset     ,
            destination_table     )                  :
        return {
            'configuration': {
                'copy': {
                    'writeDisposition': 'WRITE_TRUNCATE',
                    'destinationTable': {
                        'projectId': destination_project,
                        'datasetId': destination_dataset,
                        'tableId': destination_table,
                    },
                    'sourceTable': {
                        'projectId': source_project,
                        'datasetId': self.dataset_name,
                        'tableId': self.table_name,
                    }
                }
            }
        }

    @staticmethod
    def _make_insert_body(
            rows                      , **_3to2kwargs                                 )                  :
        insert_id_fn = _3to2kwargs['insert_id_fn']; del _3to2kwargs['insert_id_fn']
        template_suffix = _3to2kwargs['template_suffix']; del _3to2kwargs['template_suffix']
        ignore_unknown = _3to2kwargs['ignore_unknown']; del _3to2kwargs['ignore_unknown']
        skip_invalid = _3to2kwargs['skip_invalid']; del _3to2kwargs['skip_invalid']
        body = {
            'kind': 'bigquery#tableDataInsertAllRequest',
            'skipInvalidRows': skip_invalid,
            'ignoreUnknownValues': ignore_unknown,
            'rows': [{
                'insertId': insert_id_fn(row),
                'json': row,
            } for row in rows],
        }

        if template_suffix is not None:
            body['templateSuffix'] = template_suffix

        return body

    def _make_load_body(
            self, source_uris           , project     , autodetect      ,
            source_format              ,
            write_disposition             ,
            ignore_unknown_values      ,
            schema_update_options                          
    )                  :
        return {
            'configuration': {
                'load': {
                    'autodetect': autodetect,
                    'ignoreUnknownValues': ignore_unknown_values,
                    'sourceUris': source_uris,
                    'sourceFormat': source_format.value,
                    'writeDisposition': write_disposition.value,
                    'schemaUpdateOptions': [
                        e.value for e in schema_update_options],
                    'destinationTable': {
                        'projectId': project,
                        'datasetId': self.dataset_name,
                        'tableId': self.table_name,
                    },
                },
            },
        }

    def _make_query_body(
            self, query     , project     ,
            write_disposition             ,
            use_query_cache      ,
            dry_run      )                  :
        return {
            'configuration': {
                'query': {
                    'query': query,
                    'writeDisposition': write_disposition.value,
                    'destinationTable': {
                        'projectId': project,
                        'datasetId': self.dataset_name,
                        'tableId': self.table_name,
                    },
                    'useQueryCache': use_query_cache,
                },
                'dryRun': dry_run,
            },
        }

    # https://cloud.google.com/bigquery/docs/reference/rest/v2/tables/insert
    def create(self, table                ,
                     session                    = None,
                     timeout      = 60)                  :
        """Create the table specified by tableId from the dataset."""
        project = self.project()
        url = ('{}/projects/{}/datasets/'
               '{}/tables'.format((API_ROOT), (project), (self.dataset_name)))

        table['tableReference'] = {
            'projectId': project,
            'datasetId': self.dataset_name,
            'tableId': self.table_name
        }

        return self._post_json(url, table, session, timeout)

    # https://cloud.google.com/bigquery/docs/reference/rest/v2/tables/patch
    def patch(self, table                ,
                    session                    = None,
                    timeout      = 60)                  :
        """Patch an existing table specified by tableId from the dataset."""
        project = self.project()
        url = ('{}/projects/{}/datasets/'
               '{}/tables/{}'.format((API_ROOT), (project), (self.dataset_name), (self.table_name)))

        table['tableReference'] = {
            'projectId': project,
            'datasetId': self.dataset_name,
            'tableId': self.table_name
        }
        table_data = json.dumps(table).encode('utf-8')

        headers = self.headers()

        s = SyncSession(session) if session else self.session
        resp = s.patch(url, data=table_data, headers=headers,
                             timeout=timeout)
        data                 = resp.json()
        return data

    # https://cloud.google.com/bigquery/docs/reference/rest/v2/tables/delete
    def delete(self,
                     session                    = None,
                     timeout      = 60)                  :
        """Deletes the table specified by tableId from the dataset."""
        project = self.project()
        url = ('{}/projects/{}/datasets/'
               '{}/tables/{}'.format((API_ROOT), (project), (self.dataset_name), (self.table_name)))

        headers = self.headers()

        s = SyncSession(session) if session else self.session
        resp = s.session.delete(url, headers=headers, params=None,
                                      timeout=timeout)
        try:
            data                 = resp.json()
        except Exception:  # pylint: disable=broad-except
            # For some reason, `gcloud-rest` seems to have intermittent issues
            # parsing this response. In that case, fall back to returning the
            # raw response body.
            try:
                data = {'response': resp.text()}
            except (AttributeError, TypeError):
                data = {'response': resp.text}

        return data

    # https://cloud.google.com/bigquery/docs/reference/rest/v2/tables/get
    def get(
            self, session                    = None,
            timeout      = 60)                  :
        """Gets the specified table resource by table ID."""
        project = self.project()
        url = ('{}/projects/{}/datasets/'
               '{}/tables/{}'.format((API_ROOT), (project), (self.dataset_name), (self.table_name)))

        return self._get_url(url, session, timeout)

    # https://cloud.google.com/bigquery/docs/reference/rest/v2/tabledata/insertAll
    def insert(
            self, rows                      , skip_invalid       = False,
            ignore_unknown       = True, session                    = None,
            template_suffix                = None,
            timeout      = 60, **_3to2kwargs
    )                  :
        if 'insert_id_fn' in _3to2kwargs: insert_id_fn = _3to2kwargs['insert_id_fn']; del _3to2kwargs['insert_id_fn']
        else: insert_id_fn =  None
        """
        Streams data into BigQuery

        By default, each row is assigned a unique insertId. This can be
        customized by supplying an `insert_id_fn` which takes a row and
        returns an insertId.

        In cases where at least one row has successfully been inserted and at
        least one row has failed to be inserted, the Google API will return a
        2xx (successful) response along with an `insertErrors` key in the
        response JSON containing details on the failing rows.
        """
        if not rows:
            return {}

        project = self.project()
        url = ('{}/projects/{}/datasets/{}/'
               'tables/{}/insertAll'.format((API_ROOT), (project), (self.dataset_name), (self.table_name)))

        body = self._make_insert_body(
            rows, skip_invalid=skip_invalid, ignore_unknown=ignore_unknown,
            template_suffix=template_suffix,
            insert_id_fn=insert_id_fn or self._mk_unique_insert_id)
        return self._post_json(url, body, session, timeout)

    # https://cloud.google.com/bigquery/docs/reference/rest/v2/jobs/insert
    # https://cloud.google.com/bigquery/docs/reference/rest/v2/Job#jobconfigurationtablecopy
    def insert_via_copy(
            self, destination_project     , destination_dataset     ,
            destination_table     , session                    = None,
            timeout      = 60)       :
        """Copy BQ table to another table in BQ"""
        project = self.project()
        url = '{}/projects/{}/jobs'.format((API_ROOT), (project))

        body = self._make_copy_body(
            project, destination_project,
            destination_dataset, destination_table)
        response = self._post_json(url, body, session, timeout)
        return Job(response['jobReference']['jobId'], self._project,
                   session=self.session.session, token=self.token)

    # https://cloud.google.com/bigquery/docs/reference/rest/v2/jobs/insert
    # https://cloud.google.com/bigquery/docs/reference/rest/v2/Job#JobConfigurationLoad
    def insert_via_load(
            self, source_uris           , session                    = None,
            autodetect       = False,
            source_format               = SourceFormat.CSV,
            write_disposition              = Disposition.WRITE_TRUNCATE,
            timeout      = 60,
            ignore_unknown_values       = False,
            schema_update_options                                     = None
    )       :
        """Loads entities from storage to BigQuery."""
        project = self.project()
        url = '{}/projects/{}/jobs'.format((API_ROOT), (project))

        body = self._make_load_body(
            source_uris, project, autodetect, source_format, write_disposition,
            ignore_unknown_values, schema_update_options or []
        )
        response = self._post_json(url, body, session, timeout)
        return Job(response['jobReference']['jobId'], self._project,
                   session=self.session.session, token=self.token)

    # https://cloud.google.com/bigquery/docs/reference/rest/v2/jobs/insert
    # https://cloud.google.com/bigquery/docs/reference/rest/v2/Job#JobConfigurationQuery
    def insert_via_query(
            self, query     , session                    = None,
            write_disposition              = Disposition.WRITE_EMPTY,
            timeout      = 60, use_query_cache       = True,
            dry_run       = False)       :
        """Create table as a result of the query"""
        warnings.warn('using Table#insert_via_query is deprecated.'
                      'use Job#insert_via_query instead', DeprecationWarning)
        project = self.project()
        url = '{}/projects/{}/jobs'.format((API_ROOT), (project))

        body = self._make_query_body(query, project, write_disposition,
                                     use_query_cache, dry_run)
        response = self._post_json(url, body, session, timeout)
        job_id = response['jobReference']['jobId'] if not dry_run else None
        return Job(job_id, self._project,
                   session=self.session.session, token=self.token)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from future import standard_library
standard_library.install_aliases()
from pkg_resources import get_distribution
__version__ = get_distribution('gcloud-rest-bigquery').version

from gcloud.rest.bigquery.bigquery import Disposition
from gcloud.rest.bigquery.bigquery import SCOPES
from gcloud.rest.bigquery.bigquery import SchemaUpdateOption
from gcloud.rest.bigquery.bigquery import SourceFormat
from gcloud.rest.bigquery.dataset import Dataset
from gcloud.rest.bigquery.job import Job
from gcloud.rest.bigquery.table import Table
from gcloud.rest.bigquery.utils import query_response_to_dict


__all__ = [
    '__version__',
    'Dataset',
    'Disposition',
    'Job',
    'SCOPES',
    'SchemaUpdateOption',
    'SourceFormat',
    'Table',
    'query_response_to_dict',
]
